using System;
using System.IO;
using System.Windows.Forms;
using System.ComponentModel;
using System.Security;

abstract class Document
{
	String _PathName;

	enum CanSaveResult
	{
		Yes,
		No,
		SecurityError,
		AccessError
	}

	public Document()
	{
		_PathName = "";
	}

	public abstract Boolean Modified
	{
		get;
		set;
	}

	public String Title
	{
		get
		{
			if (_PathName == "")
			{
				return "Untitled";
			}
			else
			{
				return Path.GetFileName(_PathName);
			}
		}
	}

	protected abstract Boolean Read(FileStream fs);
	protected abstract Boolean Write(FileStream fs);
	protected abstract void Reset();

	Boolean ModifiedFileSavedOrDiscarded()
	{
		DialogResult result = MessageBox.Show(
			"Do you want to save the changes you made to \"" +
			this.Title +
			"\"?",
			"Save Changes",
			MessageBoxButtons.YesNoCancel,
			MessageBoxIcon.Warning
			);
		switch (result)
		{
			case DialogResult.Yes:
				// if InternalSave() fails, New() won't do anything
				return Save();
			case DialogResult.No:
				return true;
			case DialogResult.Cancel:
				break;
		}
		return false;
	}

	public Boolean New()
	{
		try
		{
			if ((this.Modified && ModifiedFileSavedOrDiscarded()) || !this.Modified)
			{
				Reset();
				_PathName = "";
				this.Modified = false;
				return true;
			}
			else
			{
				return false;
			}
		}
		catch
		{
			return false;
		}
	}

	Boolean InternalOpen(String pathName)
	{
		FileStream fs = null;
		Boolean good = true;
		try
		{
			fs = new FileStream(
				pathName,
				FileMode.Open,
				FileAccess.Read,
				FileShare.Read
				);

			good = Read(fs);

			fs.Close();
		}
		catch
		{
			good = false;
		}
		finally
		{
			if (fs != null)
			{
				fs.Close();
			}
			if (!good)
			{
				try
				{
					File.Delete(pathName);
				}
				catch
				{
				}
			}
		}
		return good;
	}

	public Boolean Open()
	{
		if ((this.Modified && ModifiedFileSavedOrDiscarded()) || !this.Modified)
		{
			OpenFileDialog ofd = new OpenFileDialog();
			ofd.Filter = "SQL Files (*.sql)|*.sql|All Files (*.*)|*.*";
			if (ofd.ShowDialog() == DialogResult.OK)
			{
				if (!InternalOpen(ofd.FileName))
				{
					ofd.Dispose();
					return false;
				}
				Reset();
				_PathName = ofd.FileName;
				this.Modified = false;
				ofd.Dispose();
				return true;
			}
			ofd.Dispose();
			return false;
		}
		else
		{
			return false;
		}
	}

	Boolean InternalSave(String pathName)
	{
		FileStream fs = null;
		Boolean good = true;
		try
		{
			fs = new FileStream(
				pathName,
				FileMode.Create,
				FileAccess.Write,
				FileShare.None
				);

			good = Write(fs);

			fs.Close();
		}
		catch
		{
			good = false;
		}
		finally
		{
			if (fs != null)
			{
				fs.Close();
			}
			if (!good)
			{
				try
				{
					File.Delete(pathName);
				}
				catch
				{
				}
			}
		}
		return good;
	}

	static CanSaveResult CanSave(String pathName)
	{
		FileStream fs = null;
		try
		{
			fs = new FileStream(
					pathName,
					FileMode.Create,
					FileAccess.Write,
					FileShare.None
					);
			fs.Close();

			return CanSaveResult.Yes;
		}
		catch (SecurityException)
		{
			return CanSaveResult.SecurityError;
		}
		catch (UnauthorizedAccessException)
		{
			return CanSaveResult.AccessError;
		}
		catch
		{
			return CanSaveResult.No;
		}
		finally
		{
			if (fs != null)
				fs.Close();
		}
	}

	public Boolean Save()
	{
		if (_PathName == "")
		{
			return SaveAs();
		}
		else
		{
			if (CanSave(_PathName) != CanSaveResult.Yes)
				return SaveAs();
			else
				return InternalSave(_PathName);
		}
	}

	public Boolean SaveAs()
	{
		SaveFileDialog sfd = new SaveFileDialog();
		sfd.FileOk += new CancelEventHandler(OnSaveFileOk);
		sfd.Filter = "SQL Files (*.sql)|*.sql|All Files (*.*)|*.*";
		sfd.FileName = this.Title;
		if (sfd.ShowDialog() == DialogResult.OK)
		{
			if (!InternalSave(sfd.FileName))
			{
				sfd.Dispose();
				return false;
			}
			_PathName = sfd.FileName;
			sfd.Dispose();
			return true;
		}
		sfd.Dispose();
		return false;
	}

	void OnSaveFileOk(Object sender, CancelEventArgs e)
	{
		try
		{
			SaveFileDialog sfd = sender as SaveFileDialog;
			e.Cancel = false;
			CanSaveResult result = CanSave(sfd.FileName);
			switch (result)
			{
				case CanSaveResult.AccessError:
					MessageBox.Show(
						"The file \"" + sfd.FileName + "\" already " +
						"exists and is read-only." + Environment.NewLine +
						"Use a different file name.",
						"Save As",
						MessageBoxButtons.OK,
						MessageBoxIcon.Warning
						);
					e.Cancel = true;
					break;
				//
				// the SaveFileDialog box takes care of locked files
				//
				// case CanSaveResult.SecurityError:
				//	MessageBox.Show(
				//		"The file \"" + sfd.FileName + "\" is already " +
				//		"in use." + Environment.NewLine +
				//		"Select a new file name or close the file that is used " +
				//		"by another program.",
				//		"Save As",
				//		MessageBoxButtons.OK,
				//		MessageBoxIcon.Warning
				//		);
				//	e.Cancel = true;
				//	break;
				//
				case CanSaveResult.No:
					e.Cancel = true;
					break;
			}
		}
		catch
		{
			e.Cancel = true;
		}
	}
}

